"""
Namer Configuration readers/verifier
"""

import configparser
import json
import os
import random
import re
import tempfile
from configparser import ConfigParser
from datetime import timedelta
from pathlib import Path

from loguru import logger
from requests_cache import BACKEND_CLASSES, BaseCache, CachedSession

from namer.configuration import NamerConfig
from namer.database import abbreviations
from namer.name_formatter import PartialFormatter


def __verify_naming_config(config: NamerConfig, formatter: PartialFormatter) -> bool:
    """
    Verifies the contents of your config file. Returns False if configuration failed.
    """
    success = True
    if not config.enable_metadataapi_genres and not config.default_genre:
        logger.error("Since enable_metadataapi_genres is not True, you must specify a default_genre")
        success = False
    success = __verify_name_string(formatter, "inplace_name", config.inplace_name) and success
    return success


def __verify_watchdog_config(config: NamerConfig, formatter: PartialFormatter) -> bool:
    """
    Verifies the contents of your config file. Returns False if configuration failed.
    """
    success = True
    if not config.enable_metadataapi_genres and not config.default_genre:
        logger.error("Since enable_metadataapi_genres is not True, you must specify a default_genre")
        success = False
    success = __verify_dir(config, "watch_dir") and success
    success = __verify_dir(config, "work_dir") and success
    success = __verify_dir(config, "failed_dir") and success
    success = __verify_dir(config, "dest_dir") and success
    success = __verify_name_string(formatter, "new_relative_path_name", config.new_relative_path_name) and success
    return success


def __verify_dir(config: NamerConfig, name: str) -> bool:
    """
    verify a config directory exist. return false if verification fails
    """
    file_name = getattr(config, name) if hasattr(config, name) else None
    if file_name and not file_name.is_dir():
        logger.error("Configured directory {}: {} is not a directory or not accessible", name, file_name)
        return False

    return True


def __verify_name_string(formatter: PartialFormatter, name: str, name_string: str) -> bool:
    """
    Verify the name format string.
    """
    values = dict(zip(formatter.supported_keys, formatter.supported_keys))
    try:
        formatter.format(name_string, values)
        return True
    except KeyError as key_error:
        logger.error("Configuration {} is not a valid file name format, please check {}", name, name_string)
        logger.error("Error message: {}", key_error)
        return False


def verify_configuration(config: NamerConfig, formatter: PartialFormatter) -> bool:
    """
    Can verify a NamerConfig with a formatter
    """
    success = __verify_naming_config(config, formatter)
    success = __verify_watchdog_config(config, formatter) and success
    return success


def from_config(config: ConfigParser) -> NamerConfig:
    """
    Given a config parser pointed at a namer.cfg file, return a NamerConfig with the file's parameters.
    """
    namer_config = NamerConfig()
    namer_config.porndb_token = config.get("namer", "porndb_token", fallback="")
    namer_config.inplace_name = config.get("namer", "inplace_name", fallback="{site} - {date} - {name}.{ext}")
    namer_config.name_parser = config.get("namer", "name_parser", fallback="{_site}{_sep}{_optional_date}{_ts}{_name}{_dot}{_ext}")
    namer_config.prefer_dir_name_if_available = config.getboolean("namer", "prefer_dir_name_if_available", fallback=False)
    namer_config.target_extensions = [x.strip().lower() for x in config.get("namer", "target_extensions", fallback="mp4,mkv,avi,mov,flv").split(",")]

    namer_config.min_file_size = config.getint("namer", "min_file_size", fallback=100)
    namer_config.set_uid = config.getint("namer", "set_uid", fallback=None)
    namer_config.set_gid = config.getint("namer", "set_gid", fallback=None)
    namer_config.trailer_location = config.get("namer", "trailer_location", fallback=None)
    namer_config.sites_with_no_date_info = [re.sub(r"[^a-z0-9]", "", x.strip().lower()) for x in config.get("namer", "sites_with_no_date_info", fallback="").split(",")]
    if "" in namer_config.sites_with_no_date_info:
        namer_config.sites_with_no_date_info.remove("")

    abbreviations_db = abbreviations.copy()
    site_abbreviations = config.get("namer", "site_abbreviations", fallback=None)
    if site_abbreviations:
        data = json.loads(site_abbreviations)
        abbreviations_db.update(data)

    new_abbreviation = {}
    for abbreviation, full in abbreviations_db.items():
        key = re.compile(fr'^{abbreviation}[ .-]+', re.IGNORECASE)
        new_abbreviation[key] = f'{full} '

    namer_config.site_abbreviations = new_abbreviation

    namer_config.override_tpdb_address = config.get("namer", "override_tpdb_address", fallback="https://api.metadataapi.net/")

    namer_config.write_namer_log = config.getboolean("namer", "write_namer_log", fallback=False)
    namer_config.write_namer_failed_log = config.getboolean("namer", "write_namer_failed_log", fallback=True)
    namer_config.update_permissions_ownership = config.getboolean("namer", "update_permissions_ownership", fallback=False)
    namer_config.set_dir_permissions = config.getint("namer", "set_dir_permissions", fallback=775)
    namer_config.set_file_permissions = config.getint("namer", "set_file_permissions", fallback=664)
    namer_config.max_performer_names = config.getint("namer", "max_performer_names", fallback=6)
    namer_config.enabled_requests_cache = config.getboolean("namer", "use_requests_cache", fallback=True)
    namer_config.requests_cache_expire_minutes = config.getint("namer", "requests_cache_expire_minutes", fallback=10)

    namer_config.movie_data_prefered = [x.strip().upper() for x in config.get("namer", "movie_data_prefered", fallback="").split(",")]
    vr_studios = "18 VR,Babe VR,Badoink VR,Dorm Room,Kink VR,Real VR,RealJamVR,Sex Like Real,SexBabesVR,SinsVR,SLR Originals,Swallowbay,Virtual Taboo,VirtualRealPorn,VR Bangers,VR Cosplay X,VR Hush,VRConk,VRedging,Wankz VR"
    namer_config.vr_studios = [x.strip().upper() for x in config.get("namer", "vr_studios", fallback=vr_studios).split(",")]
    namer_config.vr_tags = [x.strip().upper() for x in config.get("namer", "vr_tags", fallback="virtual reality, vr porn").split(",")]

    namer_config.preserve_duplicates = config.getboolean("duplicates", "preserve_duplicates", fallback=True)
    namer_config.max_desired_resolutions = config.getint("duplicates", "max_desired_resolutions", fallback=-1)
    namer_config.desired_codec = [x.strip().upper() for x in config.get("duplicates", "desired_codec", fallback="hevc,h264").split(",")]

    namer_config.write_nfo = config.getboolean("metadata", "write_nfo", fallback=False)
    namer_config.enabled_tagging = config.getboolean("metadata", "enabled_tagging", fallback=True)
    namer_config.enabled_poster = config.getboolean("metadata", "enabled_poster", fallback=True)
    namer_config.enable_metadataapi_genres = config.getboolean("metadata", "enable_metadataapi_genres", fallback=False)
    namer_config.default_genre = config.get("metadata", "default_genre", fallback="Adult")
    namer_config.language = config.get("metadata", "language", fallback=None)
    namer_config.ignored_dir_regex = config.get("metadata", "ignored_dir_regex", fallback=".*_UNPACK_.*")
    namer_config.new_relative_path_name = config.get("watchdog", "new_relative_path_name", fallback="{site} - {date} - {name}/{site} - {date} - {name}.{ext}")
    namer_config.del_other_files = config.getboolean("watchdog", "del_other_files", fallback=False)
    namer_config.extra_sleep_time = config.getint("watchdog", "extra_sleep_time", fallback=30)

    watch_dir = config.get("watchdog", "watch_dir", fallback=None)
    if watch_dir:
        namer_config.watch_dir = Path(watch_dir).resolve()

    work_dir = config.get("watchdog", "work_dir", fallback=None)
    if work_dir:
        namer_config.work_dir = Path(work_dir).resolve()

    failed_dir = config.get("watchdog", "failed_dir", fallback=None)
    if failed_dir:
        namer_config.failed_dir = Path(failed_dir).resolve()

    dest_dir = config.get("watchdog", "dest_dir", fallback=None)
    if dest_dir:
        namer_config.dest_dir = Path(dest_dir).resolve()

    namer_config.retry_time = config.get("watchdog", "retry_time", fallback=f"03:{random.randint(0, 59):0>2}")
    namer_config.web = config.getboolean("watchdog", "web", fallback=False)
    namer_config.port = config.getint("watchdog", "port", fallback=6980)
    namer_config.host = config.get("watchdog", "host", fallback="0.0.0.0")
    namer_config.web_root = config.get("watchdog", "web_root", fallback=None)
    namer_config.allow_delete_files = config.getboolean("watchdog", "allow_delete_files", fallback=False)
    namer_config.add_max_percent_column = config.getboolean("watchdog", "add_max_percent_column", fallback=False)
    namer_config.debug = config.getboolean("watchdog", "debug", fallback=False)
    namer_config.diagnose_errors = config.getboolean("watchdog", "diagnose_errors", fallback=False)

    # create a CachedSession objects for request caching.
    if namer_config.enabled_requests_cache:
        cache_file = Path(tempfile.gettempdir()) / "namer_cache"
        sqlite_supported = issubclass(BACKEND_CLASSES['sqlite'], BaseCache)
        backend = 'sqlite' if sqlite_supported else 'filesystem'
        expire_time = timedelta(minutes=namer_config.requests_cache_expire_minutes)
        namer_config.cache_session = CachedSession(str(cache_file), backend=backend, expire_after=expire_time, ignored_parameters=["Authorization"])

    return namer_config


def default_config() -> NamerConfig:
    """
    Attempts reading various locations to fine a namer.cfg file.
    """
    config = configparser.ConfigParser()
    default_locations = []
    config_loc = os.environ.get("NAMER_CONFIG")
    if config_loc and Path(config_loc).exists():
        default_locations = [Path(config_loc)]
    elif (Path.home() / ".namer.cfg").exists():
        default_locations = [Path.home() / ".namer.cfg"]
    elif Path("./namer.cfg").exists():
        default_locations = [Path("./namer.cfg")]
    config.read(default_locations)
    return from_config(config)



"""
Helper functions to tie in to namer's functionality.
"""
import gzip
import json
import math
import shutil
from functools import lru_cache
from pathlib import Path
from queue import Queue
from types import SimpleNamespace
from typing import Dict, List, Optional

import jsonpickle
from werkzeug.routing import Rule

from namer.comparison_results import ComparisonResults
from namer.configuration import NamerConfig
from namer.command import gather_target_files_from_dir, is_interesting_movie, subpath_or_equal, Command
from namer.metadataapi import __build_url, __get_response_json_object, __metadataapi_response_to_data  # type: ignore


def has_no_empty_params(rule: Rule) -> bool:
    """
    Currently unused, useful to inspect Flask rules.
    """
    defaults = rule.defaults if rule.defaults is not None else ()
    arguments = rule.arguments if rule.arguments is not None else ()
    return len(defaults) >= len(arguments)


def get_failed_files(config: NamerConfig) -> List[Dict]:
    """
    Get failed files to rename.
    """
    return list(map(lambda o: command_to_file_info(o, config), gather_target_files_from_dir(config.failed_dir, config)))


def get_queued_files(queue: Queue, config: NamerConfig, queue_limit: int = 100) -> List[Dict]:
    """
    Get queued files.
    """
    queue_items = list(queue.queue)[:queue_limit]
    return list(map(lambda x: command_to_file_info(x, config), filter(lambda i: i is not None, queue_items)))


def get_queue_size(queue: Queue) -> int:
    return queue.qsize()


def command_to_file_info(command: Command, config: NamerConfig) -> Dict:
    stat = command.target_movie_file.stat()

    res = {
        'file': str(command.target_movie_file.relative_to(command.config.failed_dir)) if subpath_or_equal(command.target_movie_file, command.config.failed_dir) else None,
        'name': command.target_directory.stem if command.parsed_dir_name and command.target_directory else command.target_movie_file.stem,
        'ext': command.target_movie_file.suffix[1:].upper(),
        'update_time': int(stat.st_mtime),
        'size': stat.st_size,
    }

    if config and config.add_max_percent_column:
        log_data = read_failed_log_file(res['file'], config)
        percentage = log_data.results[0].name_match if log_data and log_data.results and log_data.results[0] else 0.0
        res['percentage'] = percentage

    return res


def get_search_results(query: str, search_type: str, file: str, config: NamerConfig, page: int = 1) -> Dict:
    """
    Search results for user selection.
    """

    responses = {}
    if search_type == 'Any' or search_type == 'Scenes':
        # scenes
        url = __build_url(config, name=query, page=page, movie=False)
        responses[url] = __get_response_json_object(url, config)

    if search_type == 'Any' or search_type == 'Movies':
        # movies
        url = __build_url(config, name=query, page=page, movie=True)
        responses[url] = __get_response_json_object(url, config)

    file_infos = []
    for url, response in responses.items():
        if response and response.strip() != '':
            json_obj = json.loads(response, object_hook=lambda d: SimpleNamespace(**d))
            formatted = json.dumps(json.loads(response), indent=4, sort_keys=True)
            file_infos.extend(__metadataapi_response_to_data(json_obj, url, formatted, None))

    files = []
    for scene_data in file_infos:
        scene = {
            'id': scene_data.uuid,
            'type': scene_data.type,
            'title': scene_data.name,
            'date': scene_data.date,
            'poster': scene_data.poster_url,
            'site': scene_data.site,
            'tags_count': len(scene_data.tags),
            'performers': scene_data.performers,
        }
        files.append(scene)

    res = {
        'file': file,
        'files': files,
    }

    return res


def delete_file(file_name_str: str, config: NamerConfig) -> bool:
    """
    Delete selected file.
    """
    file_name = config.failed_dir / file_name_str
    if not is_acceptable_file(file_name, config) or not config.allow_delete_files:
        return False

    if config.del_other_files:
        target_name = config.failed_dir / Path(file_name_str).parts[0]
        shutil.rmtree(target_name)
    else:
        file_name.unlink()

    return not file_name.is_file()


def read_failed_log_file(name: str, config: NamerConfig) -> Optional[ComparisonResults]:
    file = config.failed_dir / name
    file = file.parent / (file.stem + '_namer.json.gz')

    print("Log file name: {}", file)

    res: Optional[ComparisonResults] = None
    if file.is_file():
        res = _read_failed_log_file(file, file.stat().st_size, file.stat().st_mtime)

    print("Res file {}", res)

    return res


@lru_cache(maxsize=1024)
def _read_failed_log_file(file: Path, file_size: int, file_update: float) -> Optional[ComparisonResults]:
    res: Optional[ComparisonResults] = None
    if file.is_file():
        data = gzip.decompress(file.read_bytes())
        decoded = jsonpickle.decode(data)
        if decoded and isinstance(decoded, ComparisonResults):
            res = decoded

    return res


def is_acceptable_file(file: Path, config: NamerConfig) -> bool:
    """
    Checks if a file belong to namer.
    """
    return str(config.failed_dir) in str(file.resolve()) and file.is_file() and is_interesting_movie(file, config)


def human_format(num):
    if num == 0:
        return '0'

    size = 1000
    size_name = ('', 'K', 'M', 'B', 'T')
    i = int(math.floor(math.log(num, size)))
    p = math.pow(size, i)
    s = str(round(num / p, 2))
    s = s.rstrip('0').rstrip('.')

    return f'{s}{size_name[i]}'



abbreviations = {
    'wgp': 'WhenGirlsPlay',
    '18og': '18OnlyGirls',
    '18yo': '18YearsOld',
    '1kf': '1000Facials',
    '21ea': '21EroticAnal',
    '21fa': '21FootArt',
    '21n': '21Naturals',
    '2cst': '2ChicksSameTime',
    'a1o1': 'Asian1on1',
    'aa': 'AmateurAllure',
    'ad': 'AmericanDaydreams',
    'add': 'ManualAddActors',
    'agm': 'AllGirlMassage',
    'am': 'AssMasterpiece',
    'analb': 'AnalBeauty',
    'baebz': 'Baeb',
    'bblib': 'BigButtsLikeItBig',
    'bcasting': 'BangCasting',
    'bconfessions': 'BangConfessions',
    'bglamkore': 'BangGlamkore',
    'bgonzo': 'BangGonzo',
    'brealteens': 'BangRealTeens',
    'bcb': 'BigCockBully',
    'bch': 'BigCockHero',
    'bdpov': 'BadDaddyPOV',
    'bex': 'BrazzersExxtra',
    'bgb': 'BabyGotBoobs',
    'bgbs': 'BoundGangbangs',
    'bin': 'BigNaturals',
    'bjf': 'BlowjobFridays',
    'bp': 'ButtPlays',
    'btas': 'BigTitsatSchool',
    'btaw': 'BigTitsatWork',
    'btc': 'BigTitCreampie',
    'btis': 'BigTitsinSports',
    'btiu': 'BigTitsinUniform',
    'btlbd': 'BigTitsLikeBigDicks',
    'btra': 'BigTitsRoundAsses',
    'burna': 'BurningAngel',
    'bwb': 'BigWetButts',
    'cfnm': 'ClothedFemaleNudeMale',
    'clip': 'LegalPorno',
    'cps': 'CherryPimps',
    'cuf': 'CumFiesta',
    'cws': 'CzechWifeSwap',
    'da': 'DoctorAdventures',
    'Daughter': 'DaughterSwap',
    'Daughters': 'DaughterSwap',
    'dbm': 'DontBreakMe',
    'dc': 'DorcelVision',
    'ddfb': 'DDFBusty',
    'ddfvr': 'DDFNetworkVR',
    'dm': 'DirtyMasseur',
    'dnj': 'DaneJones',
    'dpg': 'DigitalPlayground',
    'dwc': 'DirtyWivesClub',
    'dwp': 'DayWithAPornstar',
    'dsw': 'DaughterSwap',
    'esp': 'EuroSexParties',
    'ete': 'EuroTeenErotica',
    'ext': 'ExxxtraSmall',
    'fams': 'FamilyStrokes',
    'faq': 'FirstAnalQuest',
    'fds': 'FakeDrivingSchool',
    'fft': 'FemaleFakeTaxi',
    'fhd': 'FantasyHD',
    'fhl': 'FakeHostel',
    'fho': 'FakehubOriginals',
    'fka': 'FakeAgent',
    'fm': 'FuckingMachines',
    'fms': 'FantasyMassage',
    'frs': 'FitnessRooms',
    'ft': 'FastTimes',
    'ftx': 'FakeTaxi',
    'gft': 'GrandpasFuckTeens',
    'gbcp': 'GangbangCreampie',
    'gta': 'GirlsTryAnal',
    'gw': 'GirlsWay',
    'h1o1': 'Housewife1on1',
    'ham': 'HotAndMean',
    'hart': 'Hegre',
    'hcm': 'HotCrazyMess',
    'hegre-art': 'Hegre',
    'hoh': 'HandsOnHardcore',
    'hotab': 'HouseofTaboo',
    'ht': 'Hogtied',
    'ihaw': 'IHaveAWife',
    'iktg': 'IKnowThatGirl',
    'il': 'ImmoralLive',
    'kha': 'KarupsHA',
    'kow': 'KarupsOW',
    'kpc': 'KarupsPC',
    'la': 'LatinAdultery',
    'lcd': 'LittleCaprice',
    'lhf': 'LoveHerFeet',
    'lsb': 'Lesbea',
    'lst': 'LatinaSexTapes',
    'lta': 'LetsTryAnal',
    'maj': 'ManoJob',
    'mbb': 'MommyBlowsBest',
    'mbt': 'MomsBangTeens',
    'mc': 'MassageCreep',
    'mcu': 'MonsterCurves',
    'mdhf': 'MyDaughtersHotFriend',
    'mdhg': 'MyDadsHotGirlfriend',
    'mfa': 'ManuelFerrara',
    'mfhg': 'MyFriendsHotGirl',
    'mfhm': 'MyFriendsHotMom',
    'mfl': 'Mofos',
    'mfst': 'MyFirstSexTeacher',
    'mgbf': 'MyGirlfriendsBustyFriend',
    'mgb': 'MommyGotBoobs',
    'mic': 'MomsInControl',
    'mj': 'ManoJob',
    'mlib': 'MilfsLikeItBig',
    'mlt': 'MomsLickTeens',
    'mmgs': 'MommysGirl',
    'mnm': 'MyNaughtyMassage',
    'mom': 'MomXXX',
    'mpov': 'MrPOV',
    'mrs': 'MassageRooms',
    'mshf': 'MySistersHotFriend',
    'mts': 'MomsTeachSex',
    'mvft': 'MyVeryFirstTime',
    'mwhf': 'MyWifesHotFriend',
    'naf': 'NeighborAffair',
    'nam': 'NaughtyAmerica',
    'na': 'NaughtyAthletics',
    'naughtyamericavr': 'NaughtyAmerica',
    'nb': 'NaughtyBookworms',
    'news': 'NewSensations',
    'nf': 'NubileFilms',
    'no': 'NaughtyOffice',
    'nrg': 'NaughtyRichGirls',
    'nubilef': 'NubileFilms',
    'num': 'NuruMassage',
    'nw': 'NaughtyWeddings',
    'obj': 'OnlyBlowjob',
    'otb': 'OnlyTeenBlowjobs',
    'pav': 'PixAndVideo',
    'pba': 'PublicAgent',
    'pf': 'PornFidelity',
    'phd': 'PassionHD',
    'plib': 'PornstarsLikeitBig',
    'pop': 'PervsOnPatrol',
    'ppu': 'PublicPickups',
    'prdi': 'PrettyDirty',
    'ps': 'PropertySex',
    'pud': 'PublicDisgrace',
    'reg': 'RealExGirlfriends',
    'rkp': 'RKPrime',
    'rws': 'RealWifeStories',
    'saf': 'ShesAFreak',
    'sart': 'SexArt',
    'sbj': 'StreetBlowjobs',
    'sislove': 'SisLovesMe',
    'smb': 'ShareMyBF',
    'ssc': 'StepSiblingsCaught',
    'ssn': 'ShesNew',
    'sts': 'StrandedTeens',
    'swsn': 'SwallowSalon',
    'tdp': 'TeensDoPorn',
    'tds': 'TheDickSuckers',
    'ted': 'Throated',
    'tf': 'TeenFidelity',
    'tgs': 'ThisGirlSucks',
    'these': 'TheStripperExperience',
    'tla': 'TeensLoveAnal',
    'tlc': 'TeensLoveCream',
    'tle': 'TheLifeErotic',
    'tlhc': 'TeensLoveHugeCocks',
    'tlib': 'TeensLikeItBig',
    'tlm': 'TeensLoveMoney',
    'togc': 'TonightsGirlfriendClassic',
    'tog': 'TonightsGirlfriend',
    'tspa': 'TrickySpa',
    'tss': 'ThatSitcomShow',
    'tuf': 'TheUpperFloor',
    'wa': 'WhippedAss',
    'wfbg': 'WeFuckBlackGirls',
    'wkp': 'Wicked',
    'wlt': 'WeLiveTogether',
    'woc': 'WildOnCam',
    'wov': 'WivesOnVacation',
    'wowg': 'WowGirls',
    'wy': 'WebYoung',
    'zzs': 'ZZseries',
    'ztod': 'ZeroTolerance',
    'itc': 'InTheCrack',
    'abbw': 'AbbyWinters',
    'abme': 'AbuseMe',
    'ana': 'AnalAngels',
    'atke': 'ATKExotics',
    'atkg': 'ATKGalleria',
    'atkgfs': 'ATKGirlfriends',
    'atkh': 'ATKHairy',
    'aktp': 'ATKPetites',
    'btp': 'BadTeensPunished',
    'brealmilfs': 'Bang.RealMilfs',
    'byngr': 'bang.YNGR',
    'ba': 'Beauty-Angels',
    'bgfs': 'BlackGFS',
    'bna': 'BrandNew',
    'bam': 'BruceAndMorgan',
    'bcast': 'BrutalCastings',
    'bd': 'BrutalDildos',
    'bpu': 'BrutalPickups',
    'cfnmt': 'CFNMTeens',
    'cfnms': 'FNMSecret',
    'cza': 'CzhecAmateurs',
    'czbb': 'CzechBangBus',
    'czb': 'CzechBitch',
    'cc': 'CzechCasting',
    'czc': 'CzechCouples',
    'czestro': 'CzechEstrogenolit',
    'czf': 'CzechFantasy',
    'czgb': 'CzechGangBang',
    'cgfs': 'CzechGFS',
    'czharem': 'CzechHarem',
    'czm': 'CzechMassage',
    'czo': 'CzechOrgasm',
    'czps': 'CzechPawnShop',
    'css': 'CzechStreets',
    'cztaxi': 'CzechTaxi',
    'czt': 'CzechTwins',
    'dlla': 'DadysLilAngel',
    'dts': 'DeepThroatSirens',
    'deb': 'DeviceBondage',
    'doan': 'DiaryOfANanny',
    'dpf': 'DPFanatics',
    'ds': 'DungeonSex',
    'ffr': 'FacialsForever',
    'ff': 'FilthyFamily',
    'fbbg': 'FirstBGG',
    'fab': 'FuckedAndBound',
    'fum': 'FuckingMachines',
    'fs': 'FuckStudies',
    'tfcp': 'FullyClothedPissing',
    'gfr': 'GFRevenge',
    'gdp': 'GirlsDoPorn',
    'hletee': 'HelplessTeens',
    'hotb': 'HouseOfTaboo',
    'Infr': 'InfernalRestraints',
    'inh': 'InnocentHigh',
    'jlmf': 'JessieLoadsMonsterFacials',
    'university': 'KinkUniversity',
    'lang': 'LANewGirl',
    'mmp': 'MMPNetwork',
    'mot': 'MoneyTalks',
    'mbc': 'MyBabysittersClub',
    'mdm': 'MyDirtyMaid',
    'mfp': 'MyFamilyPies',
    'nvg': 'NetVideoGirls',
    'nubp': 'Nubiles-Porn',
    'oo': 'Only-Opaques',
    'os': 'Only-Secretaries',
    'oss': 'OnlySilAndSatin',
    'psus': 'PascalsSubSluts',
    'pbf': 'PetiteBallerinasFucked',
    'phdp': 'PetiteHDPoorn',
    'psp': 'PorsntarsPunishment',
    'pc': 'PrincessCum',
    'pdmqfo': 'QuestForOrgasm',
    'rtb': 'RealTimeBondage',
    'rab': 'RoundAndBrown',
    'sr': 'SadisticRope',
    'sas': 'SexAndSubmission',
    'sed': 'SexualDisgrace',
    'seb': 'SexuallyBroken',
    'sislov': 'SisLovesMe',
    'tslw': 'SlimeWave',
    'steps': 'StepSiblings',
    'stre': 'StrictRestraint',
    't18': 'Taboo18',
    'tft': 'TeacherFucksTeens',
    'tmf': 'TeachMeFisting',
    'tsma': 'TeenSexMania',
    'tsm': 'TeenSexMovs',
    'ttw': 'TeensInTheWoods',
    'tgw': 'ThaiGirlsWild',
    'taob': 'TheArtOfBlowJob',
    'trwo': 'TheRealWorkout',
    'tto': 'TheTrainingOfO',
    'tg': 'TopGrl',
    'tt': 'TryTeens',
    'th': 'TwistysHard',
    'vp': 'VIPissy',
    'wrh': 'WeAreHairy',
    'wpa': 'WhippedAss',
    'yt': 'YoungThroats',
    'zb': 'ZoliBoy',
}



"""
A wrapper allowing shutdown of a Flask server.
"""
import logging
import mimetypes
from datetime import datetime
from queue import Queue
from threading import Thread
from typing import List, Optional, Union

from flask import Blueprint, Flask
from flask_compress import Compress
from loguru import logger
from waitress import create_server
from waitress.server import BaseWSGIServer, MultiSocketServer

from namer.configuration import NamerConfig
from namer.web.routes import api, web


class GenericWebServer:
    """
    A wrapper allowing shutdown of a Flask server.
    """
    __app: Flask
    __compress = Compress()

    __port: int
    __host: str
    __path: str
    __blueprints: List[Blueprint]

    __server: Union[MultiSocketServer, BaseWSGIServer]
    __thread: Thread

    __mime_types: dict = {
        'font/woff': '.woff',
        'font/woff2': '.woff2',
    }

    def __init__(self, host: str, port: int, webroot: Optional[str], blueprints: List[Blueprint], static_path: Optional[str] = 'public', quiet=True):
        self.__host = host
        self.__port = port
        self.__path = '/' if not webroot else webroot
        self.__app = Flask(__name__, static_url_path=self.__path, static_folder=static_path, template_folder='templates')
        self.__blueprints = blueprints

        if quiet:
            logging.getLogger('waitress').disabled = True
            logging.getLogger('waitress.queue').disabled = True

        self.__add_mime_types()
        self.__register_blueprints()
        self.__make_server()
        self.__register_custom_processors()

    def __make_server(self):
        self.__compress.init_app(self.__app)
        self.__server = create_server(self.__app, host=self.__host, port=self.__port, clear_untrusted_proxy_headers=True)
        self.__thread = Thread(target=self.__run, daemon=True)

    def __register_blueprints(self):
        for blueprint in self.__blueprints:
            blueprint_path = self.__path + blueprint.url_prefix if blueprint.url_prefix else self.__path
            self.__app.register_blueprint(blueprint, url_prefix=blueprint_path)

    def __add_mime_types(self):
        self.__app.json.mimetype = 'application/json; charset=utf-8'  # type: ignore

        for mime, ext in self.__mime_types.items():
            test_mime, test_ext = mimetypes.guess_type(f'0{ext}')
            if test_mime is None:
                mimetypes.add_type(mime, ext)

    def __register_custom_processors(self):
        functions = {
            'bool_to_icon': self.bool_to_icon,
        }
        self.__app.jinja_env.globals.update(**functions)

        filters = {
            'timestamp_to_datetime': self.timestamp_to_datetime,
            'strftime': self.strftime,
        }
        self.__app.jinja_env.filters.update(**filters)

        self.__app.jinja_env.add_extension('jinja2.ext.do')

    def start(self):
        self.__thread.start()

    def __run(self):
        """
        Start server on existing thread.
        """
        if self.__server:
            try:
                self.__server.run()
            except OSError:
                logger.error("Stopping server")
            finally:
                self.stop()

    def stop(self):
        """
        Stop severing requests and empty threads.
        """
        if self.__server:
            self.__server.close()

    def get_effective_port(self) -> Optional[int]:
        return getattr(self.__server, "effective_port", None)

    def get_url(self) -> str:
        """
        Returns the full url to access this server, usually http://127.0.0.1:<os assigned port>/
        """
        return f"http://{self.__host}:{self.get_effective_port()}{self.__path}"

    @staticmethod
    def bool_to_icon(item: bool) -> str:
        icon = 'x'
        if item:
            icon = 'check'

        return f'<i class="bi bi-{icon}"></i>'

    @staticmethod
    def timestamp_to_datetime(item: int) -> datetime:
        return datetime.utcfromtimestamp(item)

    @staticmethod
    def strftime(item: datetime, datetime_format: str) -> str:
        return item.strftime(datetime_format)


class NamerWebServer(GenericWebServer):
    __namer_config: NamerConfig
    __command_queue: Queue

    def __init__(self, namer_config: NamerConfig, command_queue: Queue):
        self.__namer_config = namer_config
        self.__command_queue = command_queue
        webroot = "/" if not self.__namer_config.web_root else self.__namer_config.web_root
        blueprints = [
            web.get_routes(self.__namer_config, self.__command_queue),
            api.get_routes(self.__namer_config, self.__command_queue),
        ]

        super().__init__(self.__namer_config.host, self.__namer_config.port, webroot, blueprints)
